@@ -7,13 +7,15 @@ import json
 from json.encoder import JSONEncoder
 from feed.actiontypes import ReturnTypes
 from bs4 import BeautifulSoup, Tag
-from feed.settings import kafka_params, routing_params, nanny_params
-from feed.actiontypes import ActionChainException
+
 import signal
 
 
 from kafka import KafkaConsumer, KafkaProducer
 
+from feed.service import Client
+from feed.settings import kafka_params, routing_params, nanny_params
+from feed.actiontypes import ActionChainException
 
 
 class GracefulKiller:
@@ -74,7 +76,7 @@ class BrowserSearchParams(ObjectSearchParams):
                 logging.debug(f'ObjectSearchParam::_returnItem(): searching for node with attribues, class=[{cls}]')
                 out.extend(soup.findAll(attrs={'class': cls}))
             item = out
-            formatted = lambda it: it
+            formatted = lambda it: (it, it.find_parent("a").attrs if it.find_parent('a') is not None else {})
         elif self.returnType == 'attr':
             formatted = lambda element: element.get_attribute(self.attribute)
         elif self.returnType == 'element':
@@ -116,8 +118,6 @@ class BrowserSearchParams(ObjectSearchParams):
             return None
 
 
-
-
 class Action(BrowserSearchParams):
 
     def __init__(self, position, **kwargs):
@@ -136,12 +136,13 @@ class Action(BrowserSearchParams):
             logging.debug(f'Action::execute: Action executed succesfully, name=[{chain.name}], position=[{action.position}]')
             return ret
         except ActionChainException as ex:
-            Action.publishActionEror(ex)
+            Action.publishActionError(chain, ex)
             logging.info(f'{type(ex).__name__} thrown whilst processing')
             return False
         except Exception as ex:
             traceback.print_exc()
             logging.warning(f'Action::execute:: {type(ex).__name__} thrown whilst processing name=[{chain.name}], position=[{action.position}], args=[{ex.args}]')
+            Action.publishUnhandledActionError(chain, ex, action)
             return False
             # TODO Exception reporting callback called here
             # OnClickException for example
@@ -156,21 +157,29 @@ class Action(BrowserSearchParams):
         # TODO For UI-Server
         return [self.__dict__().keys()]
 
-    def publishActionEror(actionException):
-        requests.put('http://{host}:{port}/actionsmanager/reportActionError/{name}'.format(name=actionException.chainName, **nanny_params), json=actionException.__dict__())
+    @staticmethod
+    def publishActionError(chain, actionException):
+        actionException.userID = chain.nannyClient.behalf
+        chain.nannyClient.put(f'/actionsmanager/reportActionError/{actionException.chainName}', payload=actionException.__dict__())
+
+    @staticmethod
+    def publishUnhandledActionError(chain, exception, action):
+        #chain.nannyClient.put(f'')
+        logging.warning(f'Unhandled exception in action {action.__dict__()}, exception={exception.args}')
+        pass
 
 
 class CaptureAction(Action):
     def __init__(self, **kwargs):
         super().__init__(**kwargs)
-        self.kwargs.pop('data', None)
-        self.kwargs.pop('src', None)
-        self.returnType = 'src'
-        self.captureName = kwargs['captureName'] # mandatory
-        self.data = kwargs.get('data', None)
+        self.returnType = self.kwargs.pop('returnType', 'src')
+        self.parentAttributes = self.kwargs.pop('parentAttributes', {})
+        self.captureName = self.kwargs.pop('captureName') # mandatory
+        self.data = self.kwargs.pop('data', None)
+        self.backupKey = self.kwargs.pop('backupKey', None)
 
     def __dict__(self):
-        return dict(data=self.data, **self.kwargs)
+        return dict(data=self.data, parentAttributes=self.parentAttributes, backupKey=self.backupKey, captureName=self.captureName, returnType=self.returnType, **self.kwargs)
 
 class InputAction(Action):
 
@@ -197,6 +206,7 @@ class ClickAction(Action):
         self.isSingle = True
         self.returnType = 'element'
 
+
 ActionTypes = {
     "ClickAction": ClickAction,
     "InputAction": InputAction,
@@ -204,6 +214,7 @@ ActionTypes = {
     "PublishAction": PublishAction
 }
 
+
 class ActionChain:
     actions= {}
 
@@ -212,7 +223,12 @@ class ActionChain:
         self.name = kwargs.get('name')
         self.startUrl = kwargs.get('startUrl')
         self.repeating = kwargs.get('isRepeating', True)
+        self.userID = kwargs.get('userID', None)
         actionParams = kwargs.get('actions', [])
+        self.isSample = False
+
+        self.nannyClient = Client('nanny', behalf=self.userID, check_health=False, **nanny_params)
+        self.routerClient = Client('router', behalf=self.userID, check_health=False, **routing_params)
         self.failedChain = False
         for order, params in enumerate(actionParams):
             try:
@@ -223,21 +239,29 @@ class ActionChain:
                 traceback.print_exc()
                 logging.error(f'{type(self).__name__}::__init__(): chainName=[{self.name}], position=[{order}] actionType=[{params.get("actionType")}] is missing {ex.args} default parameter')
 
-    def recoverHistory(self):
-        try:
-            req = requests.get('http://{host}:{port}/routingcontroller/getLastPage/{name}'.format(name=self.name, **routing_params))
-        except Exception as e:
-            logging.warning(f'ActionChain::recoverHistory: router is unavailable')
-        logging.info(f'{type(self).__name__}::recoverHistory have {req} from routing.')
-        try:
-            data = req.json()
-        except Exception:
-            logging.warning(f'Did not get valid response from router. response=[{req}]')
-        return data
+    def __repr__(self):
+        return f'{type(self).__name__}: name={self.name}'
 
-    def saveHistory(self, url):
+    def recoverHistory(self):
+        req = self.routerClient.get(f'/routingcontroller/getLastPage/{self.name}', resp=True, errors=[])
+        logging.info(f'{type(self).__name__}::recoverHistory have {req} from routing.')
+        return req
+
+    def saveHistory(self):
         pass
 
+    def shouldRun(self):
+        logging.debug(f'Checking if {self.name} should run')
+        if self.failedChain:
+            return False
+        for actionIndex in self.actions:
+            logging.debug(f'Checking if {actionIndex} in {self.name} can be run')
+            errors = self.nannyClient.get(f'/actionsmanager/findActionErrorReports/{self.name}/{actionIndex}', resp=True, error=[])
+            if len(errors) > 0:
+                logging.info(f'Will not run {self.name}')
+                return False
+        return True
+
     @staticmethod
     def actionFactory(position, actionParams):
         logging.info(f'ActionChain::actionFactory: node=[{position}]: {", ".join(map(lambda key: "{}=[{}]".format(key, actionParams[key]), actionParams))}')
@@ -265,7 +289,13 @@ class ActionChain:
         """
         logging.info(f'{type(self).__name__}::onChainEnd(): chainName={self.name}')
 
+    def initialiseClients(self):
+        logging.info(f'Initialising clients with {self.userID}')
+        self.nannyClient.behalf = self.userID
+        self.routerClient.behalf = self.userID
+
     def execute(self, caller, initialise=True):
+        self.initialiseClients()
         self.failedChain = False
         if initialise:
             self.initialise(caller)
@@ -280,11 +310,18 @@ class ActionChain:
                 continue
             callBackMethod = getattr(caller, f'on{type(action).__name__}Callback')
             for item in success:
-                callBackMethod(item)
+                try:
+                    callBackMethod(item, chain=self)
+                except ActionChainException as ex:
+                    logging.warning(f'{type(ex).__name__} raised during on{type(action).__name__}CallBack')
+                    ex.chainName = self.name
+                    Action.publishActionError(self, ex)
             self.saveHistory()
             self.onChainEnd()
 
     def getRepublishRoute(self, action):
+        logging.debug(f'Getting sample route for {action}. isSample={self.isSample}')
+        sample = lambda route: f'{route}-sample'
         name = 'unknown'
         if isinstance(action, CaptureAction):
             name = 'summarizer-route'
@@ -295,7 +332,10 @@ class ActionChain:
             name = 'leader-route'
         route = f'{os.getenv("KAFKA_TOPIC_PREFIX", "d")}-{name}'
         logging.info(f'republishing to route=[{route}]')
-        return route
+        if self.isSample:
+            return sample(route)
+        else:
+            return route
 
     def rePublish(self, action, *args, **kwargs):
         pass
@@ -322,20 +362,20 @@ class KafkaActionPublisher(ActionChain):
         self.producer = KafkaProducer(**kafka_params, value_serializer=lambda m: bytes(json.dumps(m, cls=ActionReturnSerialiser).encode('utf-8')))
         self.messages_out_since_flush = 0
 
-    def rePublish(self, actionReturn):
+    def rePublish(self, actionReturn, topic):
         self.messages_out_since_flush += 1
-        topic = self.getRepublishRoute(actionReturn.action)
         # construct chain parameters to send
         actionReturn.action.data = actionReturn.data
         payload = {
             'actions': [actionReturn.action],
             'startUrl': actionReturn.current_url,
             'isRepeating': False,
-            'name': actionReturn.name
+            'name': actionReturn.name,
+            'userID': actionReturn.userID
         }
 
         self.producer.send(topic=topic, value=payload)
-        logging.info(f'{type(self).__name__}::rePublish(): Republished ActionReturn(startUrl={actionReturn.current_url} for {type(actionReturn.action).__name__} to {topic}')
+        logging.info(f'{type(self).__name__}: Republished ActionReturn(startUrl={actionReturn.current_url} for {type(actionReturn.action).__name__} to {topic}')
         # check to see if we should flush
         if self.messages_out_since_flush >= self.flush_rate:
             self.messages_out_since_flush = 0
@@ -380,36 +420,34 @@ class ActionChainRunner:
     def main(self):
         killer = GracefulKiller()
         logging.info(f'{type(self).__name__}::main(): beginning subscription poll of kafka')
-        for actionChainParams in self.subscription():
+        for actionChainParams, route in self.subscription():
+            # If shutdown was triggered, then do that now
+            if killer.kill_now:
+                break
             if not self.driverHealthCheck():
                 self.renewDriverSession()
-            if killer.kill_now:
-                self.cleanUp()
-                logging.info(f'cleaned up resources')
-                break
-            if not self.shouldRun(actionChain):
-                logging.info(f'Skipping {actionChainParams.get("name")}.')
+
+            actionChain = self.implementation(driver=self.driver, **actionChainParams)
+            if 'sample' in route:
+                actionChain.isSample = True
+            if not actionChain.shouldRun():
+                logging.info(f'Skipping {actionChain.name}.')
                 # TODO notifications service here
                 continue
-            actionChain = self.implementation(driver=self.driver, **actionChainParams)
             logging.info(f'{type(self).__name__}::main(): START:{actionChain.name} implementing action chain {actionChainParams.get("name")}: {json.dumps(actionChainParams, indent=4)}')
             ret = actionChain.execute(self)
             self.onChainEndCallback(actionChain, ret)
             # should the chain be automatically be re ran from where we are?
             # this can be disabled in the implementation of onChainEndCallback
-            while actionChain.repeating and not actionChain.failedChain and self.shouldRun(actionChain):
+            while actionChain.repeating and actionChain.shouldRun():
+                if killer.kill_now:
+                    break
                 ret = actionChain.execute(caller=self, initialise=False)
                 self.onChainEndCallback(actionChain, ret)
+            if killer.kill_now:
+                break
             logging.info(f'{type(self).__name__}::main(): END:{actionChain.name} ActionChain::execute() has returned')
-
-    def shouldRun(self, actionChain: ActionChain):
-        for actionIndex in range(actionChain.actions):
-            logging.debug(f'Checking if {actionIndex} in {actionChain.name} can be run')
-            errorReq = requests.get('http://{host}:{port}/actionsmanager/findActionErrorReports/{name}/{pos}'.format(**nanny_params, name=actionChain.name, pos=actionIndex))
-            errors = errorReq.json()
-            if len(errors) > 0:
-                logging.info(f'Will not run {actionChain.name}')
-                return False
+        self.cleanUp()
 
     def cleanUp(self):
         logging.warning(f'ActionChainRunner::cleanUp() No cleanup has been implemented')
@@ -417,17 +455,26 @@ class ActionChainRunner:
 
 class KafkaActionSubscription(ActionChainRunner):
 
-    def __init__(self, topic,  **kwargs):
+    def __init__(self, *topics,  **kwargs):
         super().__init__(**kwargs)
-        logging.info(f'Starting ActionChainRuner type {type(self).__name__}, topic=[{topic}], prefix=[{os.environ["KAFKA_TOPIC_PREFIX"]}]')
-        self._consumer = KafkaConsumer(**kafka_params, value_deserializer=lambda m: json.loads(m.decode('utf-8')))
-        self.topic = f'{os.environ["KAFKA_TOPIC_PREFIX"]}-{topic}'
+        self.topics = list(map(KafkaActionSubscription.topic_name, topics))
 
+        logging.info(f'Starting ActionChainRuner type {type(self).__name__}, topics=[{self.topics}], prefix=[{os.environ["KAFKA_TOPIC_PREFIX"]}]')
+        self._consumer = KafkaConsumer(**kafka_params, value_deserializer=lambda m: json.loads(m.decode('utf-8')))
+
+    @staticmethod
+    def topic_name(topic):
+        return  f'{os.environ["KAFKA_TOPIC_PREFIX"]}-{topic}'
+
+    @staticmethod
+    def get_route(message):
+        return message.topic.split('{os.environ["KAFKA_TOPIC_PREFIX"]}-')[-1]
 
     def subscription(self):
-        self._consumer.subscribe([self.topic])
+        self._consumer.subscribe(self.topics)
         for mes in self._consumer:
-            yield mes.value
+            route = KafkaActionSubscription.get_route(mes)
+            yield mes.value, route
 
 
 class CommandsActionSubscription(ActionChainRunner):
