@@ -1,12 +1,11 @@
-from feed.logger import getLogger
 from queue import Queue
+import logging
 import os
 import sys
 import traceback
 from http.client import RemoteDisconnected
 from time import sleep, time
 import re
-import requests as r
 import selenium.webdriver as webdriver
 from bs4 import BeautifulSoup
 from bs4 import Tag, NavigableString
@@ -14,23 +13,19 @@ from selenium.webdriver.chrome.options import Options
 from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
 from selenium.webdriver.remote.webdriver import WebDriver
 from selenium.webdriver.remote.webelement import WebElement
-from selenium.common.exceptions import WebDriverException
+from selenium.common.exceptions import WebDriverException, NoSuchElementException
 from urllib3.exceptions import MaxRetryError, ProtocolError
-import requests
 import threading
 import subprocess
 import argparse
 
-from feed.settings import nanny_params, browser_params, routing_params
+from feed.settings import browser_params
 from feed.service import Client
 
 from feed.actionchains import ObjectSearchParams, ActionChain, ClickAction, InputAction, CaptureAction, PublishAction, Action
 from feed.actiontypes import ActionableItemNotFound
 
 
-logging = getLogger(__name__)
-
-
 def verifyUrl(url):
     if url is None:
         return False
@@ -51,17 +46,20 @@ class BrowserActions(ActionChain):
             self.current_url = current_url
             self.data = data
             self.action = action
+            self.userID = kwargs.get('userID')
 
         def __dict__(self):
-            return dict(name=self.name, current_url=self.current_url, data=self.data, action=self.action.__dict__())
+            return dict(name=self.name, userID=self.userID, current_url=self.current_url, data=self.data, action=self.action.__dict__())
 
     driver = None # type: WebDriver
 
     def __init__(self, driver: WebDriver, *args, **kwargs):
         super().__init__(*args, **kwargs)
-        requests.get('http://{host}:{port}/routingcontroller/initialiseRoutingSession/{name}'.format(name=self.name, **routing_params))
+        self.routerClient.get(f'/routingcontroller/initialiseRoutingSession/{self.name}')
         self.kwargs = kwargs
         self.driver = driver
+        self.soup = BeautifulSoup("<div>None</div>")
+        self.backupKeyIncrement = 0
 
     @staticmethod
     def _get_button_to_click(item, action):
@@ -85,6 +83,10 @@ class BrowserActions(ActionChain):
         else:
             return None
 
+    def _update_soup(self):
+        logging.info(f'Updating soup for {self}')
+        self.soup = BeautifulSoup(self.driver.page_source)
+
     @staticmethod
     def _searchNavigableStringForTag(navString: NavigableString, text):
         if navString == text:
@@ -95,7 +97,6 @@ class BrowserActions(ActionChain):
         else:
             return None
 
-
     def _verify_class_string(self, item):
 
         if ' ' in item:
@@ -129,10 +130,27 @@ class BrowserActions(ActionChain):
                     logging.debug(f'checking class names, classNames=[{classNames}], len=[{len(classNames)}]')
                     for className in classNames:
                         elems = self.driver.find_elements_by_class_name(className)
-                        logging.info(f'found {len(elems)} with className=[{className}]')
+                        logging.info(f'found {len(elems)} elements with className=[{className}]')
                         if len(elems) == 1:
                             logging.info(f'{type(self).__name__}: found unique button to click with className=[{className}], should only have appeared here once')
                             button = elems[0]
+                        else:
+                            found = False
+                            logging.info(f'Checking elems={list(map(lambda item: item.text, elems))}')
+                            for item in elems:
+                                try:
+                                    it = item.find_element_by_link_text(action.text)
+                                except NoSuchElementException:
+                                    if action.text.upper() in str(item).upper():
+                                        it = item
+                                    else:
+                                        continue
+                                if it:
+                                    logging.info(f'Found by text {it}, {action.text}')
+                                    button = it
+                                    found = True
+                                    break
+
         logging.info(f'{type(self).__name__}::onClickAction(): clicking on text={button.text}')
         clickingFrom = self.driver.current_url
         clickTime = time()
@@ -145,17 +163,20 @@ class BrowserActions(ActionChain):
             else:
                 logging.debug(f'{type(self).__name__}::onClickAction(): current url has changed from "{clickingFrom}" to "{self.driver.current_url}"')
             logging.info(f'{type(self).__name__}::onClickAction(): current url has changed from "{clickingFrom}" to "{self.driver.current_url}"')
-        return [BrowserActions.Return(current_url=self.driver.current_url, name=self.name, action=action, data=None)]
+        self._update_soup()
+        return [BrowserActions.Return(current_url=self.driver.current_url, userID=self.userID, name=self.name, action=action, data=None)]
 
     def onCaptureAction(self, action: CaptureAction):
+        action.backupKey = f'{self.driver.current_url}'
         data = action.getActionableItem(action, self.driver)
         if data is None:
             raise ActionableItemNotFound(position=action.position, actionHash=action.getActionHash(), chainName=self.name)
         self.rePublish(key=self.driver.current_url, action=action, data=data)
         if not action.isSingle:
-            return [BrowserActions.Return(current_url=self.driver.current_url, name=self.name, action=action,data=item) for item in data]
+            return [BrowserActions.Return(current_url=self.driver.current_url, userID=self.userID, name=self.name, action=action, data=item) for item in data]
         else:
-            return [BrowserActions.Return(current_url=self.driver.current_url, name=self.name, action=action,data=data)]
+            logging.debug(f'returning data={data}, action={action}')
+            return [BrowserActions.Return(current_url=self.driver.current_url, userID=self.userID, name=self.name, action=action,data=data)]
 
     def onPublishAction(self, action: PublishAction):
         logging.info(f'{type(self).__name__}::onPublishAction: css=[{action.css}], xpath=[{action.xpath}], text=[{action.text}]')
@@ -212,24 +233,28 @@ class BrowserActions(ActionChain):
                 out.append(child.attrs.get('href'))
         if len(out) == 0:
             raise ActionableItemNotFound(position=action.position, actionHash=action.getActionHash(), chainName=self.name)
-        # TODO should put this into a callback
-        # TODO should republish chain here
+        # TODO should put rePublishing into a callback.
         self.rePublish(key=self.driver.current_url, action=action, data=out)
         if not action.isSingle:
-            return [BrowserActions.Return(current_url=self.driver.current_url, name=self.name, action=action,data=url) for url in out]
+            return [BrowserActions.Return(current_url=self.driver.current_url, userID=self.userID, name=self.name, action=action,data=url) for url in out]
         else:
-            return [BrowserActions.Return(current_url=self.driver.current_url, name=self.name, action=action,data=out[0])]
+            return [BrowserActions.Return(current_url=self.driver.current_url, name=self.name, userID=self.userID, action=action,data=out[0])]
 
     def onInputAction(self, action: InputAction):
         inputField: WebElement = action.getActionableItem(action, self.driver)
-        inputField.send_keys(action.inputString)
+        try:
+            inputField.send_keys(action.inputString)
+        except AttributeError as ex:
+            # TODO should also handle element not interactable here and give invalid element found back to user
+            # TODO in the above case, should try surrounding elements - general rule should be to go inwards
+            raise ActionableItemNotFound(position=action.position, actionHash=action.getActionHash(), chainName=self.name)
         logging.warning(f'{type(self).__name__}::onInputAction: chain=[{self.name}], action=[{action}]')
-        return [BrowserActions.Return(current_url=self.driver.current_url, name=self.name, action=action, data=inputField)]
+        return [BrowserActions.Return(current_url=self.driver.current_url, name=self.name, action=action, userID=self.userID, data=inputField)]
 
     def saveHistory(self):
         try:
             logging.info(f'BrowserActions::saveHistory: Saving current_url=[{self.driver.current_url}]')
-            requests.get('http://{host}:{port}/routingcontroller/updateHistory/{name}'.format(name=self.name, **routing_params), data=self.driver.current_url)
+            self.routerClient.get(f'/routingcontroller/updateHistory/{self.name}', payload=self.driver.current_url)
         except Exception as e:
             logging.warning(f'BrowserActions::saveHistory: router is unavailable.')
 
@@ -248,9 +273,10 @@ class BrowserActions(ActionChain):
             caller.renewDriverSession()
             self.driver = caller.driver
             self.driver.get(url)
+        self._update_soup()
 
-        ret = BrowserActions.Return(action=None, data=None, current_url=self.driver.current_url, name=self.name)
-        caller.initialiseCallback(ret)
+        ret = BrowserActions.Return(action=None, data=None, userID=self.userID, current_url=self.driver.current_url, name=self.name)
+        caller.initialiseCallback(ret, chain=self)
 
 
 class BrowserService:
@@ -331,19 +357,20 @@ class BrowserService:
         """
         open a process, forward its logs and watch for commands in a queue.
         """
-        logging.info(f'BrowserService::beginBrowserThread(): Starting web browser thread')
+        sellogger = logging.getLogger('crawling.SeleniumProcessLogger')
+        sellogger.info(f'BrowserService::beginBrowserThread(): Starting web browser thread')
         with subprocess.Popen(os.getenv('SELENIUM_PROCESS_SCRIPT',  "/opt/bin/start-selenium-standalone.sh"), stdout=subprocess.PIPE, bufsize=1, universal_newlines=True) as process:
             for line, command in BrowserService._bind_queue_and_log(self.browser_process_command_queue, process.stdout):
-                logging.info(f'BrowserService:: {line}')
+                sellogger.info(f'BrowserService:: {line}')
                 if command is None:
-                    logging.debug(f'nothing in queue')
+                    sellogger.debug(f'nothing in queue')
                     continue
                 elif command == 'KILL':
-                    logging.debug(f'have queue item')
-                    logging.info(f'BrowserService::beginBrowserThread(): Killing process {process.id}')
+                    sellogger.debug(f'have queue item')
+                    sellogger.info(f'BrowserService::beginBrowserThread(): Killing process {process.id}')
                     process.kill()
-                    logging.debug(f'monitor thread is {"alive" if self.browser_monitor_thread.is_alive() else "complete"}')
-        logging.info(f'Browser process {process.pid} has been torn down.')
+                    sellogger.debug(f'monitor thread is {"alive" if self.browser_monitor_thread.is_alive() else "complete"}')
+        sellogger.info(f'Browser process {process.pid} has been torn down.')
 
     def _browser_clean_up(self):
         """
@@ -356,12 +383,6 @@ class BrowserService:
 
 
 def reportParameter(parameter_key=None):
-    endpoint = "http://{host}:{port}/parametermanager/reportParameter/{}/{}/{}".format(
-        os.getenv("NAME"),
-        parameter_key,
-        "leader",
-        **nanny_params
-    )
-    r.get(endpoint)
+    pass
 
 
